local utf8 = require 'utf8'

local array = require 'array'
local my_utf8 = require 'my_utf8'
local t = require 'utils'
local rects = require 'rects'
local Loc = require 'loc'
local move = require 'move'

local edit = {}
local I = {}  -- for file-local helpers, so I can refer to them before defining them
edit.internal = I

local _  -- idiom for unused variables

-- some constants people might like to tweak
local Background_color = {r=1, g=1, b=1}
local Text_color = {r=0, g=0, b=0}
local Cursor_color = {r=1, g=0, b=0}
local Highlight_color = {r=0.7, g=0.7, b=0.9, a=0.4}  -- selected text

local Screen_width, Screen_height, Screen_flags = nil
local Font_height = require 'font_height'

function edit.is_this_love_version_supported()
  local major_version, minor_version = love.getVersion()
  local version = major_version..'.'..minor_version
  local supported_versions = {
    ['11.5']=true,
    ['11.4']=true,
    ['12.0']=true,
  }
  return supported_versions[version]
end

function edit.preferred_love_version()
  return '11.5'
end

local Current_time, Last_focus_time, Last_resize_time, Cursor_time

-- initialize editor
-- this should happen only once regardless of how many editors you create
function edit.load()
  -- for hysteresis in a few places
  Current_time = 0
  Last_focus_time = 0  -- https://love2d.org/forums/viewtopic.php?p=249700
  Last_resize_time = 0

  -- blinking cursor (I assume you never show the cursor in more than one editor)
  Cursor_time = 0
end

function edit.new_from_defaults(top, left, right, bottom)
  local font = love.graphics.newFont(Font_height)
  -- initialize window width/height and make window resizable
  if Screen_width == nil then
    Screen_width, Screen_height, Screen_flags = love.window.getMode()
    Screen_flags.resizable = true
    love.window.setMode(Screen_width, Screen_height, Screen_flags)
  end
  return edit.new(top, left, right, bottom, font, Font_height)
end

function edit.new_from_settings(settings, top, left, right, bottom)
  local font = love.graphics.newFont(settings.font_height)
  -- set up desired window dimensions and make window resizable
  _, _, Screen_flags = love.window.getMode()
  Screen_flags.resizable = true
  Screen_width, Screen_height = settings.width, settings.height
  love.window.setMode(Screen_width, Screen_height, Screen_flags)
  local os = love.system.getOS()
  if os == 'Linux' then
    -- love.window.setPosition doesn't quite seem to do what is asked of it on Linux.
    love.window.setPosition(settings.x, settings.y-37, settings.displayindex)
  else
    love.window.setPosition(settings.x, settings.y, settings.displayindex)
  end
  local editor = edit.new(top, left, right, bottom, font, settings.font_height)
  edit.load_file(editor, settings.filename)
  editor.screen_top = settings.screen_top
  editor.cursor = settings.cursor
  if not I.valid_loc(editor, editor.screen_top)
       or not I.valid_loc(editor, editor.cursor) then
    I.scroll_to_top(editor)
  end
  return editor
end

function edit.resize(editor, w, h, right, bottom)
  Screen_width, Screen_height = w, h
  editor.right = right
  editor.width = editor.right-editor.left
  editor.bottom = bottom
  Last_resize_time = Current_time
end

function edit.load_file(editor, filename)
  editor.filename = filename
  local infile = I.open_for_reading(editor.filename)
  editor.lines = {}
  if infile then
    for line in infile:lines() do
      table.insert(editor.lines, line)
    end
    infile:close()
  end
  if #editor.lines == 0 then
    table.insert(editor.lines, '')
  end
  I.scroll_to_top(editor)  -- clear out stale locations
end

function edit.on_focus(in_focus)
  if in_focus then
    Last_focus_time = Current_time
  end
end

function edit.update(editor, dt)
  Current_time = Current_time + dt
  -- some hysteresis while resizing
  if Current_time < Last_resize_time + 0.1 then
    return
  end
  --
  Cursor_time = Cursor_time + dt
end

function edit.new(top, left, right, bottom, font, font_height)
  local result = {
    top = math.floor(top),
    left = math.floor(left),
    right = math.floor(right),
    bottom = math.floor(bottom),
    width = right-left,

    font = font,
    font_height = font_height,
    line_height = math.floor(font_height*1.3),

    -- The editor is for editing an array of lines.
    -- The array of lines can never be empty; there must be at least one line for positioning a cursor at.
    lines = {''},  -- array of strings

    -- We need to track a couple of _locations_:
    screen_top = {line=1, pos=1},  -- location at top of screen, to start drawing from
    cursor = {line=1, pos=1},  -- for search and selections
    selection1 = {},

    filename = nil,

    -- undo
    history = {},
    next_history = 1,

    -- search
    search_term = nil,
    search_backup = nil,  -- stuff to restore when cancelling search
  }
  return result
end

function edit.draw(editor, fg)
  local cursor_or_mouse_loc = editor.cursor
  if editor.mouse_down then
    cursor_or_mouse_loc = I.loc_at_mouse(editor)
  end
  local y = editor.top
--?   local mx,my = love.mouse.getPosition()
  for line_index, line in array.each(editor.lines, editor.screen_top.line) do
    local rect
    if line_index == editor.screen_top.line then
      rect = rects.compute(editor, editor.screen_top)
    else
      rect = rects.compute(editor, {line=line_index, pos=1})
    end
    for _,s in ipairs(rect.screen_line_rects) do
      t.set_color_from_table(s.bg or Background_color)
      love.graphics.rectangle('fill', editor.left+s.x, y+s.y, s.dx,s.dy)
--?       love.graphics.setColor(0.8, 0.8, 0.8, 0.1)
--?       love.graphics.rectangle('fill', editor.left+s.x,y+s.y, s.dx,s.dy-5)
      for _,c in ipairs(s.char_rects) do
--?         love.graphics.setColor(0.8, 0.9, 0.8, 0.1)
--?         love.graphics.rectangle('fill', editor.left+c.x,y+c.y, c.dx-2,c.dy-1)
        if c.data and not c.conceal then
          t.set_color_from_table(c.bg or s.bg or Background_color)
          love.graphics.rectangle('fill', editor.left+c.x, y+c.y, c.dx,c.dy)
        end
        if c.show_cursor then  -- only the primary place where the cursor can be is available for highlight
          if I.in_selection(editor, line_index, c.pos, cursor_or_mouse_loc)
              or I.in_search(editor, line_index, c.pos)
          then
            t.set_color_from_table(Highlight_color)
            love.graphics.rectangle('fill', editor.left+c.x, y+c.y, c.dx,c.dy)
          end
        end
        if c.draw then
          for _,s in ipairs(c.draw) do I.draw_shape(s, editor.left, y, fg or Text_color) end
        end
        if c.data and not c.conceal then
          t.set_color_from_table(c.fg or s.fg or fg or Text_color)
          love.graphics.print(c.data, editor.left+c.x, y+c.y)
        end
      end
    end
    y = y + rect.dy
    if y + editor.line_height > editor.bottom then
      break
    end
  end
  if editor.search_term then
    I.draw_search_bar(editor)
  end
end

function edit.on_mouse_pressed(editor, mx,my, mouse_button)
  Cursor_time = 0  -- ensure cursor is visible immediately after it moves
  if editor.search_term then return end
  editor.mouse_down = mouse_button
  local loc = move.to_loc(editor, mx,my)
  if loc then
    editor.selection1 = loc
  end
end

function edit.on_mouse_released(editor, mx,my, mouse_button)
  Cursor_time = 0  -- ensure cursor is visible immediately after it moves
  if editor.search_term then return end
  editor.mouse_down = nil
  local loc = move.to_loc(editor, mx,my)
  if Loc.eq(loc, editor.selection1) then
    editor.selection1 = {}
  else
    editor.cursor = loc
  end
end

function edit.on_mouse_wheel_moved(editor, dx,dy)
  Cursor_time = 0  -- ensure cursor is visible immediately after it moves
  if dy > 0 then
    for i=1,math.floor(dy) do
      move.up_arrow(editor)
    end
  elseif dy < 0 then
    for i=1,math.floor(-dy) do
      move.down_arrow(editor)
    end
  end
end

function I.scroll_to_top(editor)
  assert(#editor.lines > 0)
  editor.screen_top = {line=1, pos=1}
  editor.cursor = {line=1, pos=1}
end

function I.valid_loc(editor, loc)
  assert(#editor.lines > 0)
  if loc == nil then return end
  if loc.line > #editor.lines then return end
  return true
end

---- keyboard handling

local Modifiers = {'lctrl', 'rctrl', 'lalt', 'ralt', 'lshift', 'rshift', 'lgui', 'rgui'}

-- love.keyboard.isDown doesn't work on Android, so emulate it using
-- keypressed and keyreleased events
local Keys_down = {}

function edit.on_key_pressed(editor, key, scancode, isrepeat)
  Keys_down[key] = true  -- TODO: don't repeat this for multiple editors
  if array.find(Modifiers, key) then
    -- do nothing when the modifier is pressed
    return
  end
  -- include the modifier(s) when the non-modifer is pressed
  I.on_keychord_pressed(editor, I.combine_modifiers(key), key)
end

function edit.on_key_released(editor, key, scancode)
  Keys_down[key] = nil
  I.on_key_released(editor, key, scancode)
end

function edit.on_text_input(editor, t)
  -- ignore events for some time after window in focus (mostly alt-tab)
  if Current_time < Last_focus_time + 0.01 then
    return
  end
  --
  Cursor_time = 0  -- ensure cursor is visible immediately after it moves
  if love.mouse.isDown(1) then return end
  if editor.search_term then
    editor.search_term = editor.search_term..t
    I.search_next(editor)
  end
end

function I.on_keychord_pressed(editor, chord, key)
  -- ignore events for some time after window in focus (mostly alt-tab)
  if Current_time < Last_focus_time + 0.01 then
    return
  end
  --
  Cursor_time = 0  -- ensure cursor is visible immediately after it moves
  if editor.search_term then
    if chord == 'escape' then
      editor.search_term = nil
      editor.cursor = editor.search_backup.cursor
      editor.screen_top = editor.search_backup.screen_top
      editor.search_backup = nil
    elseif chord == 'return' then
      editor.search_term = nil
      editor.search_backup = nil
    elseif chord == 'backspace' then
      local len = utf8.len(editor.search_term)
      local byte_offset = my_utf8.offset(editor.search_term, len)
      editor.search_term = string.sub(editor.search_term, 1, byte_offset-1)
      editor.cursor = t.deepcopy(editor.search_backup.cursor)
      editor.screen_top = t.deepcopy(editor.search_backup.screen_top)
      I.search_next(editor)
    elseif chord == 'down' then
      if editor.cursor.pos <= utf8.len(editor.lines[editor.cursor.line]) then
        editor.cursor.pos = editor.cursor.pos+1
      else
        editor.cursor = {line=(editor.cursor.line+1)%#editor.lines, pos=1}
      end
      I.search_next(editor)
    elseif chord == 'up' then
      I.search_previous(editor)
    end
    return
--?   elseif chord == 'C-f' then
--?     editor.search_term = ''
--?     editor.search_backup = {
--?       cursor=t.deepcopy(editor.cursor),
--?       screen_top=t.deepcopy(editor.screen_top),
--?     }
  -- zoom
  elseif chord == 'C-=' then
    edit.update_font_settings(editor, editor.font_height+2)
  elseif chord == 'C--' then
    if editor.font_height > 2 then
      edit.update_font_settings(editor, editor.font_height-2)
    end
  elseif chord == 'C-0' then
    edit.update_font_settings(editor, 20)
  -- clipboard
  elseif chord == 'C-a' then
    editor.selection1 = {line=1, pos=1}
    editor.cursor = {line=#editor.lines, pos=utf8.len(editor.lines[#editor.lines])+1}
  elseif chord == 'C-c' then
    local s = I.selection(editor)
    if s then
      love.system.setClipboardText(s)
    end
  --== shortcuts that move the cursor
  elseif chord == 'up' then
    move.up_arrow(editor)
    editor.selection1 = {}
  elseif chord == 'down' then
    move.down_arrow(editor)
    editor.selection1 = {}
  elseif chord == 'pageup' then
    move.pageup(editor)
    editor.selection1 = {}
  elseif chord == 'pagedown' then
    move.pagedown(editor)
    editor.selection1 = {}
  end
end

function I.on_key_released(editor, key, scancode)
  -- ignore events for some time after window in focus (mostly alt-tab)
  if Current_time < Last_focus_time + 0.01 then
    return
  end
  --
  Cursor_time = 0  -- ensure cursor is visible immediately after it moves
end

function edit.update_font_settings(editor, font_height)
  editor.font_height = font_height
  editor.font = love.graphics.newFont(editor.font_height)
  editor.line_height = math.floor(font_height*1.3)
end

function I.draw_text_cursor(editor, x, y, dy)
  -- blink every 0.5s
  if math.floor(Cursor_time*2)%2 == 0 then
    t.set_color_from_table(Cursor_color)
    love.graphics.rectangle('fill', x,y, 3,dy)
  end
end

---- helpers for keyboard handling

function I.combine_modifiers(key)
  if love.keyboard.isModifierActive then  -- waiting for LÖVE v12
    if key:match('^kp') then
      key = I.translate_numlock(key)
    end
  end
  local result = ''
  if I.ctrl_down() then
    result = result..'C-'
  end
  if I.alt_down() then
    result = result..'M-'
  end
  if I.shift_down() then
    result = result..'S-'  -- don't try to use this with letters/digits
  end
  if I.cmd_down() then
    result = result..'s-'
  end
  result = result..key
  return result
end

function I.any_modifier_down()
  return I.ctrl_down() or I.alt_down() or I.shift_down() or I.cmd_down()
end

function I.ctrl_down()
  return Keys_down.lctrl or Keys_down.rctrl
end

function I.alt_down()
  return Keys_down.lalt or Keys_down.ralt
end

function I.shift_down()
  return Keys_down.lshift or Keys_down.rshift
end

function I.cmd_down()
  return Keys_down.lgui or Keys_down.rgui
end

function I.is_cursor_movement(key)
  return array.find({'left', 'right', 'up', 'down', 'home', 'end', 'pageup', 'pagedown'}, key)
end

-- mappings only to non-printable keys; leave out mappings that textinput will handle
local Numlock_off = {
  kp0='insert',
  kp1='end',
  kp2='down',
  kp3='pagedown',
  kp4='left',
  -- numpad 5 translates to nothing
  kp6='right',
  kp7='home',
  kp8='up',
  kp9='pageup',
  ['kp.']='delete',
  -- LÖVE handles keypad operators in textinput
  -- what's with the `kp=` and `kp,` keys? None of my keyboards have one.
  -- Hopefully LÖVE handles them as well in textinput.
  kpenter='enter',
  kpdel='delete',
}
local Numlock_on = {
  kpenter='enter',
  kpdel='delete',
}
function I.translate_numlock(key)
  if love.keyboard.isModifierActive('numlock') then
    return Numlock_on[key] or key
  else
    return Numlock_off[key] or key
  end
  return key
end

function I.draw_shape(shape, left, top, fg)
  t.set_color_from_table(shape.fg or fg)
  if shape.type == 'rect' then
    love.graphics.rectangle(shape.mode, left+shape.x, top+shape.y, shape.dx, shape.dy)
  elseif shape.type == 'line' then
    love.graphics.line(left+shape.x1, top+shape.y1, left+shape.x2, top+shape.y2)
  end
end

-- helpers for file operations

function I.open_for_reading(filename)
  local result = love.filesystem.newFile(filename)
  local ok, err = result:open('r')
  assert(ok)
  return result
end

-- helpers for the search bar (C-f)

function I.draw_search_bar(editor)
  local h = editor.line_height+2
  local y = Screen_height-h
  love.graphics.setColor(0.9,0.9,0.9)
  love.graphics.rectangle('fill', 0, y-10, Screen_width-1, h+8)
  love.graphics.setColor(0.6,0.6,0.6)
  love.graphics.line(0, y-10, Screen_width-1, y-10)
  love.graphics.setColor(1,1,1)
  love.graphics.rectangle('fill', 20, y-6, Screen_width-40, h+2, 2,2)
  love.graphics.setColor(0.6,0.6,0.6)
  love.graphics.rectangle('line', 20, y-6, Screen_width-40, h+2, 2,2)
  t.set_color_from_table(Text_color)
  love.graphics.print(editor.search_term, 25,y-5)
  I.draw_text_cursor(editor, 25+editor.font:getWidth(editor.search_term),y-5, editor.line_height)
end

function I.search_next(editor)
  local offset
  -- search current line from cursor
  local curr_pos = editor.cursor.pos
  local curr_line = editor.lines[editor.cursor.line]
  local curr_offset = my_utf8.offset(curr_line, curr_pos)
  offset = I.find(curr_line, editor.search_term, curr_offset, --[[literal]] true)
  if offset then
    editor.cursor.pos = utf8.len(curr_line, 1, offset)
  end
  if offset == nil then
    -- search lines below cursor
    for i=editor.cursor.line+1,#editor.lines do
      local curr_line = editor.lines[i]
      offset = I.find(curr_line, editor.search_term, --[[from start]] nil, --[[literal]] true)
      if offset then
        editor.cursor = {line=i, pos=utf8.len(curr_line, 1, offset)}
        break
      end
    end
  end
  if offset == nil then
    -- wrap around
    for i=1,editor.cursor.line-1 do
      local curr_line = editor.lines[i]
      offset = I.find(curr_line, editor.search_term, --[[from start]] nil, --[[literal]] true)
      if offset then
        editor.cursor = {line=i, pos=utf8.len(curr_line, 1, offset)}
        break
      end
    end
  end
  if offset == nil then
    -- search current line until cursor
    local curr_line = editor.lines[editor.cursor.line]
    offset = I.find(curr_line, editor.search_term, --[[from start]] nil, --[[literal]] true)
    local pos = utf8.len(curr_line, 1, offset)
    if pos and pos < editor.cursor.pos then
      editor.cursor.pos = pos
    end
  end
  if offset then
    move.maybe_snap_cursor_to_bottom_of_screen(editor)
  else
    -- roll back
    editor.cursor = t.deepcopy(editor.search_backup.cursor)
    editor.screen_top = t.deepcopy(editor.search_backup.screen_top)
  end
end

function I.search_previous(editor)
  local offset
  -- search current line before cursor
  local curr_pos = editor.cursor.pos
  local curr_line = editor.lines[editor.cursor.line]
  local curr_offset = my_utf8.offset(curr_line, curr_pos)
  offset = I.rfind(curr_line, editor.search_term, curr_offset-1, --[[literal]] true)
  if offset then
    editor.cursor.pos = utf8.len(curr_line, 1, offset)
  end
  if offset == nil then
    -- search lines above cursor
    for i=editor.cursor.line-1,1,-1 do
      local curr_line = editor.lines[i]
      offset = I.rfind(curr_line, editor.search_term, --[[from end]] nil, --[[literal]] true)
      if offset then
        editor.cursor = {line=i, pos=utf8.len(curr_line, 1, offset)}
        break
      end
    end
  end
  if offset == nil then
    -- wrap around
    for i=#editor.lines,editor.cursor.line+1,-1 do
      local curr_line = editor.lines[i]
      offset = I.rfind(curr_line, editor.search_term, --[[from end]] nil, --[[literal]] true)
      if offset then
        editor.cursor = {line=i, pos=utf8.len(curr_line, 1, offset)}
        break
      end
    end
  end
  if offset == nil then
    -- search current line after cursor
    local curr_line = editor.lines[editor.cursor.line]
    offset = I.rfind(curr_line, editor.search_term, --[[from end]] nil, --[[literal]] true)
    local pos = utf8.len(curr_line, 1, offset)
    if pos and pos > editor.cursor.pos then
      editor.cursor.pos = pos
    end
  end
  if offset then
    move.maybe_snap_cursor_to_top_of_screen(editor)
  else
    -- roll back
    editor.cursor = t.deepcopy(editor.search_backup.cursor)
    editor.screen_top = t.deepcopy(editor.search_backup.screen_top)
  end
end

-- return true if cursor marks start of search term and line_index,pos is in that region
function I.in_search(editor, line_index, pos)
  if editor.search_term == nil then return false end
  if #editor.search_term == 0 then return false end
  if line_index ~= editor.cursor.line then return false end
  return I.find_at(editor.lines[line_index], editor.search_term, editor.cursor.pos)
    and editor.cursor.pos <= pos and pos <= editor.cursor.pos+utf8.len(editor.search_term)-1
end

function I.find_at(data, pat, pos)
  local offset = utf8.offset(data, pos)
  return data:sub(offset, offset+#pat-1) == pat
end

function I.find(s, pat, i, plain)
  if s == nil then return end
  return s:find(pat, i, plain)
end

-- TODO: avoid the expensive reverse() operations
-- Particularly if we only care about literal matches, we don't need all of string.find
function I.rfind(s, pat, i, plain)
  if s == nil then return end
  if #pat == 0 then return #s end
  local rs = s:reverse()
  local rpat = pat:reverse()
  if i == nil then i = #s end
  local ri = #s - i + 1
  local rendpos = rs:find(rpat, ri, plain)
  if rendpos == nil then return nil end
  local endpos = #s - rendpos + 1
  assert (endpos >= #pat, ('rfind: endpos %d should be >= #pat %d at this point'):format(endpos, #pat))
  return endpos-#pat+1
end

--? function test_rfind()
--?   check_eq(I.rfind('abc', ''), 3, 'empty pattern')
--?   check_eq(I.rfind('abc', 'c'), 3, 'final char')
--?   check_eq(I.rfind('acbc', 'c', 3), 2, 'previous char')
--?   check_nil(I.rfind('abc', 'd'), 'missing char')
--?   check_nil(I.rfind('abc', 'c', 2), 'no more char')
--? end

-- helpers for selecting portions of text

function I.loc_at_mouse(editor)
  local mx,my = love.mouse.getPosition()
  return move.to_loc(editor, mx, my)
end

function I.cut_selection(editor)
  if editor.selection1.line == nil then return end
  local result = I.selection(editor)
  I.delete_selection_and_record_undo_event(editor)
  return result
end

function I.delete_selection_and_record_undo_event(editor)
  if editor.selection1.line == nil then return end
  local minl,maxl = t.minmax(editor.selection1.line, editor.cursor.line)
  local before = I.snapshot(editor, minl, maxl)
  I.delete_selection_without_undo(editor)
  I.record_undo_event(editor, {before=before, after=I.snapshot(editor, editor.cursor.line)})
end

function I.delete_selection_without_undo(editor)
  if editor.selection1.line == nil then return end
  -- min,max = sorted(editor.selection1,editor.cursor)
  local minl,minp = editor.selection1.line,editor.selection1.pos
  local maxl,maxp = editor.cursor.line,editor.cursor.pos
  if minl > maxl then
    minl,maxl = maxl,minl
    minp,maxp = maxp,minp
  elseif minl == maxl then
    if minp > maxp then
      minp,maxp = maxp,minp
    end
  end
  -- update editor.cursor and editor.selection1
  editor.cursor.line = minl
  editor.cursor.pos = minp
  if Loc.lt(editor.cursor, editor.screen_top) then
    editor.screen_top = I.loc_hor(editor, editor.cursor, editor.left)
  end
  editor.selection1 = {}
  -- delete everything between min (inclusive) and max (exclusive)
  local min_offset = my_utf8.offset(editor.lines[minl], minp)
  local max_offset = my_utf8.offset(editor.lines[maxl], maxp)
  if minl == maxl then
    editor.lines[minl] = editor.lines[minl]:sub(1, min_offset-1)..editor.lines[minl]:sub(max_offset)
    return
  end
  assert(minl < maxl, ('minl %d not < maxl %d'):format(minl, maxl))
  local rhs = editor.lines[maxl]:sub(max_offset)
  for i=maxl,minl+1,-1 do
    table.remove(editor.lines, i)
  end
  editor.lines[minl] = editor.lines[minl]:sub(1, min_offset-1)..rhs
end

function I.selection(editor)
  if editor.selection1.line == nil then return end
  -- min,max = sorted(editor.selection1,editor.cursor)
  local minl,minp = editor.selection1.line,editor.selection1.pos
  local maxl,maxp = editor.cursor.line,editor.cursor.pos
  if minl > maxl then
    minl,maxl = maxl,minl
    minp,maxp = maxp,minp
  elseif minl == maxl then
    if minp > maxp then
      minp,maxp = maxp,minp
    end
  end
  local min_offset = my_utf8.offset(editor.lines[minl], minp)
  local max_offset = my_utf8.offset(editor.lines[maxl], maxp)
  if minl == maxl then
    return editor.lines[minl]:sub(min_offset, max_offset-1)
  end
  assert(minl < maxl, ('minl %d not < maxl %d'):format(minl, maxl))
  local result = {editor.lines[minl]:sub(min_offset)}
  for i=minl+1,maxl-1 do
    table.insert(result, editor.lines[i])
  end
  table.insert(result, editor.lines[maxl]:sub(1, max_offset-1))
  return table.concat(result, '\n')
end

function I.in_selection(editor, line_index, pos, cursor)
  if editor.selection1.line == nil then return false end
  local curr = {line=line_index, pos=pos}
  if Loc.eq(cursor, editor.selection1) then
    return false
  elseif Loc.lt(cursor, editor.selection1) then
    return Loc.le(cursor, curr) and Loc.lt(curr, editor.selection1)
  elseif Loc.lt(editor.selection1, cursor) then
    return Loc.le(editor.selection1, curr) and Loc.lt(curr, cursor)
  end
end

-- undo/redo by managing the sequence of events in the current session
-- based on https://github.com/akkartik/mu1/blob/master/edit/012-editor-undo.mu

-- makes a copy of lines on every single keystroke; will be inefficient with really long lines.
-- TODO: highlight stuff inserted by any undo/redo operation
-- TODO: coalesce multiple similar operations

function I.record_undo_event(editor, data)
  editor.history[editor.next_history] = data
  editor.next_history = editor.next_history+1
  for i=editor.next_history,#editor.history do
    editor.history[i] = nil
  end
end

function I.undo_event(editor)
  if editor.next_history > 1 then
--?     print('moving to history', editor.next_history-1)
    editor.next_history = editor.next_history-1
    local result = editor.history[editor.next_history]
    return result
  end
end

function I.redo_event(editor)
  if editor.next_history <= #editor.history then
--?     print('restoring history', editor.next_history+1)
    local result = editor.history[editor.next_history]
    editor.next_history = editor.next_history+1
    return result
  end
end

-- Copy all relevant global editor.
-- Make copies of objects; the rest of the app may mutate them in place, but undo requires immutable histories.
function I.snapshot(editor, s,e)
  -- Snapshot everything by default, but subset if requested.
  assert(s, 'failed to snapshot operation for undo history')
  if e == nil then
    e = s
  end
  assert(#editor.lines > 0, 'failed to snapshot operation for undo history')
  if s < 1 then s = 1 end
  if s > #editor.lines then s = #editor.lines end
  if e < 1 then e = 1 end
  if e > #editor.lines then e = #editor.lines end
  -- compare with edit.new
  local event = {
    screen_top=t.deepcopy(editor.screen_top),
    selection=t.deepcopy(editor.selection1),
    cursor=t.deepcopy(editor.cursor),
    lines={},
    start_line=s,
    end_line=e,
    -- no filename; undo history is cleared when filename changes
  }
  for i=s,e do
    table.insert(event.lines, t.deepcopy(editor.lines[i]))
  end
  return event
end

function I.patch(lines, from, to)
--?   if #from.lines == 1 and #to.lines == 1 then
--?     assert(from.start_line == from.end_line)
--?     assert(to.start_line == to.end_line)
--?     assert(from.start_line == to.start_line)
--?     lines[from.start_line] = to.lines[1]
--?     return
--?   end
  assert(from.start_line == to.start_line, 'failed to patch undo operation')
  for i=from.end_line,from.start_line,-1 do
    table.remove(lines, i)
  end
  assert(#to.lines == to.end_line-to.start_line+1, 'failed to patch undo operation')
  for i=1,#to.lines do
    table.insert(lines, to.start_line+i-1, to.lines[i])
  end
end

return edit
