local edit = require 'edit'
local t = require 'utils'
require 'widgets'

local Axes = {
  -- fractal
  {'hilbert', 'fibonacci'},
  -- positioning
  {'1pass', '2pass'},
  -- slider
  {'noslider', 'slider'},
}

local Titles = {
  -- fractal: hilbert
  {
    -- positioning: none
    {
      'Hilbert curve',
      'Hilbert curve interactive',
    },
    -- positioning: smart
    {
      'Hilbert curve with smart positioning',
      'Hilbert curve interactive with smart positioning',
    },
  },
  -- fractal: fibonacci
  {
    -- positioning: none
    {
      'Fibonacci curve',
      'Fibonacci curve interactive',
    },
    -- positioning: smart
    {
      'Fibonacci curve with smart positioning',
      'Fibonacci curve interactive with smart positioning',
    },
  },
}

local Current_program = {1, 1, 1}
local Compare_program = nil

local Current_key = nil
local Current_key_at = nil
local Current_key_alpha = nil

local Mode = 'run'  -- or 'diff'
local Default_diff_color = {r=0.5, g=0.5, b=0.5}
local Next_mode

local Screen_width, Screen_height = 0, 0

local Margin_top = 15
local Margin_left = 25
local Margin_right = 50
local Gutter = 20
local Line_width = 40

local Title_font_height = 20
local Title_font
local Title_bar_height = Title_font_height + Margin_top
local Toolbar_height = 40

local Current_time = 0
local Cube_rotation = 0

local Editor
local Editor2  -- for comparing

function love.load(arg)
  if not edit.is_this_love_version_supported() then
    local major_version, minor_version = love.getVersion()
    error(
      ("This app hasn't been tested with LÖVE version %d.%d. Please add %d.%d to `is_this_love_version_supported` in edit.lua to try it out, and switch to version %s if you run into issues."):format(major_version, minor_version, major_version, minor_version, edit.preferred_love_version()))
  end

  love.keyboard.setKeyRepeat(true)

  love.graphics.setBackgroundColor(1,1,1)

  Title_font = love.graphics.newFont(Title_font_height)

  edit.load()

  local screen_width, screen_height = love.window.getMode()
  Screen_width, Screen_height = love.window.getMode()
  Editor = edit.new_from_defaults(Margin_top + Title_bar_height + Toolbar_height, Margin_left, screen_width-Margin_right, screen_height-Margin_top)
  Editor.right = math.min(Editor.left + Editor.font:getWidth('m')*Line_width, screen_width-Margin_right)
  Editor.width = Editor.right - Editor.left
  load_primary_program()

  Editor2 = edit.new_from_defaults(Margin_top + Title_bar_height + Toolbar_height, Editor.right + Gutter, screen_width-Margin_right, screen_height-Margin_top)
  Editor2.right = Editor2.left + Editor2.font:getWidth('m')*Line_width
  Editor2.width = Editor2.right - Editor2.left

  love.window.setTitle('curio1.love')
end

function love.resize(w,h)
  Screen_width, Screen_height = w, h
  local right = math.min(Editor.left + Editor.font:getWidth('m')*Line_width, w-Margin_right)
  edit.resize(Editor, w, h, right, h-Margin_top)
  Editor2.left = Editor.right + Gutter
  right = Editor2.left + Editor2.font:getWidth('m')*Line_width
  edit.resize(Editor2, w, h, right, h-Margin_top)
  clear_widgets()
  if Mode == 'run' then
    initialize_editor_toolbar(Editor, Current_program)
  else
    initialize_editor_diff_toolbar(Editor, Current_program)
    initialize_editor_diff_toolbar(Editor2, Compare_program)
  end
end

function love.draw()
--?   for k in pairs(widgets) do print(k) end
  if Mode == 'run' then
    draw_title(Editor, Current_program)
    love.graphics.setFont(Editor.font)
    edit.draw(Editor, Default_text_color)
    if app.draw then app.draw() end
  elseif Mode == 'diff' then
    draw_title(Editor, Current_program)
    draw_title(Editor2, Compare_program)
    draw_editor_border(Editor)
    draw_editor_border(Editor2)
    love.graphics.setFont(Editor.font)
    edit.draw(Editor, Default_diff_color)
    edit.draw(Editor2, Default_diff_color)
  end
  draw_cube(Screen_width - 160, Screen_height - 160, 80)
  widgets.__draw()
  draw_current_key()
end

function draw_title(editor, program)
  love.graphics.setColor(0,0,0, 0.4)
  love.graphics.setFont(Title_font)
  love.graphics.print(program_title(program), editor.left+10, Margin_top)
end

function draw_editor_border(editor)
  love.graphics.setLineWidth(1)
  love.graphics.setColor(0.8, 0.8, 0.8)
  love.graphics.rectangle('line', editor.left-5, editor.top-5, editor.width+10, editor.bottom - editor.top + 10, 5,5)
end

function draw_x_line(x, y)
  love.graphics.line(x, y+Toolbar_height/2-5, x+Toolbar_height/2, y+Toolbar_height/2-5)
end

function draw_y_line(x, y)
  love.graphics.line(x, y+5, x, y+5+Toolbar_height/2)
end

function draw_z_line(x, y)
  local zf = 0.45
  love.graphics.line(x, y+7, x + Toolbar_height*zf, y+7 + Toolbar_height*zf)
end

function draw_cube(x0, y0, d)
  love.graphics.setLineWidth(3)
  love.graphics.push()
  love.graphics.translate(x0,y0)
  local zf = 0.4  -- hacky foreshortening
  love.graphics.translate(d/2+zf*d, d/2+zf*d)
  love.graphics.rotate(Cube_rotation)
  love.graphics.translate(-d/2, -d/2)
  local i = 0
  local shimmer = 0.1
  for x = 0,1 do
    for y = 0,1 do
      love.graphics.setColor(0,0.4+i*shimmer,0, 0.1)
      i = (i+1)%4
      love.graphics.line(x*d, y*d, (1-x)*d, y*d)
      love.graphics.line(x*d, y*d, x*d, (1-y)*d)
      love.graphics.setColor(0,0.4+i*shimmer,0, 0.1)
      i = (i+1)%4
      love.graphics.line(x*d+zf*d, y*d+zf*d, (1-x)*d+zf*d, y*d+zf*d)
      love.graphics.line(x*d+zf*d, y*d+zf*d, x*d+zf*d, (1-y)*d+zf*d)
      love.graphics.setColor(0,0.4+i*shimmer,0, 0.1)
      i = (i+1)%4
      love.graphics.line(x*d, y*d, x*d+zf*d, y*d+zf*d)
      love.graphics.line(x*d, y*d, x*d+zf*d, y*d+zf*d)
    end
  end
  -- draw current point on the cube
  love.graphics.setColor(0,0.4,0, 0.4)
  if Mode == 'run' then
    local x, y, z = unpack(Current_program)
    love.graphics.circle('fill', (x-1)*d + zf*(z-1)*d, (y-1)*d + zf*(z-1)*d, 5)
  elseif Mode == 'diff' then
    local x, y, z = unpack(Current_program)
    local x1, y1 = (x-1)*d + zf*(z-1)*d, (y-1)*d + zf*(z-1)*d
    x, y, z = unpack(Compare_program)
    local x2, y2 = (x-1)*d + zf*(z-1)*d, (y-1)*d + zf*(z-1)*d
    love.graphics.setLineWidth(5)
    love.graphics.line(x1,y1, x2,y2)
  end
  love.graphics.pop()
end

function love.mousepressed(x,y, mouse_button)
  love.keyboard.setTextInput(true)  -- bring up keyboard on touch screen
  edit.on_mouse_pressed(Editor, x,y, mouse_button)
  edit.on_mouse_pressed(Editor2, x,y, mouse_button)
  widgets.__mousepressed(x,y, b)
end

function love.mousereleased(x,y, mouse_button)
  edit.on_mouse_released(Editor, x,y, mouse_button)
  edit.on_mouse_released(Editor2, x,y, mouse_button)
  widgets.__mousereleased(x,y, b)
end

function love.wheelmoved(dx,dy)
  edit.on_mouse_wheel_moved(Editor, dx,dy)
  edit.on_mouse_wheel_moved(Editor2, dx,dy)
end

function love.keypressed(key, scancode, isrepeat)
  Current_key = key
  Current_key_at = Current_time
  Current_key_alpha = 0.5
  edit.on_key_pressed(Editor, key, scancode, isrepeat)
  edit.on_key_pressed(Editor2, key, scancode, isrepeat)
end

function load_primary_program()
  local filename = 'alt/'..program_name(Current_program)
  edit.load_file(Editor, filename)
  clear_widgets()
  execute_primary_program()
  initialize_editor_toolbar(Editor, Current_program)
end

function clear_widgets()
  if widgets then
    for k,v in pairs(widgets) do
      if k:find('__') ~= 1 then
        widgets[k] = nil
      end
    end
  end
end

function execute_primary_program()
  app = {}
  local f = love.filesystem.load(Editor.filename)
  f()
end

function switch_program(index)
  Current_program = adjacent_program(Current_program, index)
  load_primary_program()
end

function initialize_editor_toolbar(editor, program)
  local x = editor.left-5
  local y = Margin_top+Title_bar_height
  local x2 = x
  widgets.draw_toolbar = null_button(function()
    love.graphics.setColor(0, 0, 0, 0.1)  -- transparent background; widgets get drawn in any order
    love.graphics.rectangle('fill', x2, y, editor.width+10, Toolbar_height-10)
  end)
  x = x+5
  local x2 = x
  widgets.switch_text = null_button(function()
    love.graphics.setColor(0,0,0, 0.3)
    love.graphics.print('switch:', x2, y+5)
  end)
  x = x+100
  widgets.switch_x = switch_button(1, draw_x_line, x,y, Toolbar_height/2, Toolbar_height/2)
  x = x+Toolbar_height/2+10
  widgets.switch_y = switch_button(2, draw_y_line, x,y, Toolbar_height/2, 0)
  x = x+Toolbar_height/2+10
  local zf = 0.45
  widgets.switch_z = switch_button(3, draw_z_line, x,y, Toolbar_height/2, Toolbar_height*zf)
  x = x+100
  local x2 = x
  widgets.compare_text = null_button(function()
    love.graphics.setColor(0,0,0, 0.3)
    love.graphics.print('compare:', x2, y+5)
  end)
  love.graphics.setColor(0,0,1)
  x = x+120
  widgets.compare_x = compare_button(program, 1, draw_x_line, x,y, Toolbar_height/2, Toolbar_height/2)
  x = x+Toolbar_height/2+10
  widgets.compare_y = compare_button(program, 2, draw_y_line, x,y, Toolbar_height/2, 0)
  x = x+Toolbar_height/2+10
  local zf = 0.45
  widgets.compare_z = compare_button(program, 3, draw_z_line, x,y, Toolbar_height/2, Toolbar_height*zf)
  draw_z_line(x, y)
  x = x+100
end

function null_button(draw_icon)
  return {draw=draw_icon, ispress = function() return false end}
end

function switch_button(index, draw_icon, x, y, w, inner_w)
  local draw = function()
    love.graphics.setLineWidth(3)
    love.graphics.setColor(0,0,1)
    draw_icon(x+(w-inner_w)/2, y)
  end
  local ispress = function(x2,y2)
    return x2 >= x and x2 <= x+w and y2 >= y and y2 <= y+Toolbar_height
  end
  local press = function()
    switch_program(index)
  end
  return {draw=draw, ispress=ispress, press=press}
end

function adjacent_program(program, index)
  local result = t.deepcopy(program)
  result[index] = 3 - result[index]
  return result
end

function program_name(program)
  return ('%s-%s-%s'):format(Axes[1][program[1]], Axes[2][program[2]], Axes[3][program[3]])
end

function program_title(program)
  local x, y, z = unpack(program)
  return Titles[x][y][z]
end

function love.textinput(t)
  Current_key = t
  Current_key_at = Current_time
  Current_key_alpha = 0.5
  edit.on_text_input(Editor, t)
  edit.on_text_input(Editor2, t)
end

function load_diff_program(editor, axis, program)
  local filename = 'alt/'..diff_program_name(axis, program)
  edit.load_file(editor, filename)
  initialize_editor_diff_toolbar(editor, program)
end

function compare_program(index)
  app = {}
  clear_widgets()
  Mode = 'diff'
  Current_program[index] = 1
  load_diff_program(Editor, index, Current_program)
  Compare_program = t.deepcopy(Current_program)
  Compare_program[index] = 2
  load_diff_program(Editor2, index, Compare_program)
end

function initialize_editor_diff_toolbar(editor, program)
  local x = editor.left-5
  local y = Margin_top+Title_bar_height
  local x2 = x
  widgets[editor.filename..'.draw_toolbar'] = null_button(function()
    love.graphics.setColor(0, 0, 0, 0.1)  -- transparent background; widgets get drawn in any order
    love.graphics.rectangle('fill', x2, y, editor.width+10, Toolbar_height-10)
  end)
  x = x+10
  local x2 = x
  widgets[editor.filename..'.run_text'] = run_button(program, x,y, Toolbar_height/2, function()
    love.graphics.setColor(0,0,1)
    love.graphics.print('run', x2, y+5)
  end)
  x = x+100
  local x2 = x
  widgets[editor.filename..'.compare_text'] = null_button(function()
    love.graphics.setColor(0,0,0, 0.3)
    love.graphics.print('compare:', x2, y+5)
  end)
  x = x+120
  widgets[editor.filename..'.compare_x'] = compare_button(program, 1, draw_x_line, x,y, Toolbar_height/2, Toolbar_height/2)
  x = x+Toolbar_height/2+20
  widgets[editor.filename..'.compare_y'] = compare_button(program, 2, draw_y_line, x,y, Toolbar_height/2, 0)
  x = x+Toolbar_height/2+10
  local zf = 0.45
  widgets[editor.filename..'.compare_z'] = compare_button(program, 3, draw_z_line, x,y, Toolbar_height/2, Toolbar_height*zf)
  x = x+100
end

function run_button(program, x,y, w, draw_icon)
  local draw = function()
    love.graphics.setLineWidth(3)
    love.graphics.setColor(0,0,1)
    draw_icon()
  end
  local ispress = function(x2,y2)
    return x2 >= x and x2 <= x+w and y2 >= y and y2 <= y+Toolbar_height
  end
  local press = function()
    Mode, Next_mode = 'run', nil
    Current_program = program
    load_primary_program()
  end
  return {draw=draw, ispress=ispress, press=press}
end

function compare_button(program, index, draw_icon, x, y, w, inner_w)
  local draw = function()
    love.graphics.setLineWidth(3)
    love.graphics.setColor(0,0,1)
    draw_icon(x+(w-inner_w)/2, y)
  end
  local ispress = function(x2,y2)
    return x2 >= x and x2 <= x+w and y2 >= y and y2 <= y+Toolbar_height
  end
  local press = function()
    Current_program = program
    compare_program(index)
  end
  return {draw=draw, ispress=ispress, press=press}
end

local Remaining_axes = {
  {2, 3},
  {1, 3},
  {1, 2},
}

function diff_program_name(axis, program)
  local r = Remaining_axes[axis]
  return ('%s-%s/%s'):format(Axes[r[1]][program[r[1]]], Axes[r[2]][program[r[2]]], program_name(program))
end

function love.keyreleased(key, scancode)
  edit.on_key_released(Editor, key, scancode)
  edit.on_key_released(Editor2, key, scancode)
end

function love.update(dt)
  Current_time = Current_time + dt
  if Current_key_at then
    if Current_time - Current_key_at > 0.5 then
      Current_key, Current_key_at, Current_key_alpha = nil
    else
      Current_key_alpha = (0.5 - Current_time + Current_key_at)
    end
  end
  if Current_time % 4 > 2 then
    Cube_rotation = Cube_rotation + 1e-3*math.random(15)/10
  else
    Cube_rotation = Cube_rotation - 1e-3*math.random(15)/10
  end
  edit.update(Editor, dt)
  widgets.__update(dt)
end

function love.focus(in_focus)
  edit.on_focus(in_focus)
end

-- just for demo videos
function draw_current_key()
  if not Current_key then return end
  local y = Screen_height-30-10
  local w = love.graphics.getFont():getWidth(Current_key) + 10
  local x = (Screen_width-w-10)/2
  love.graphics.setColor(0.7,0.7,0.7, Current_key_alpha)
  love.graphics.rectangle('fill', x, y, w, 30, 5,5)
  love.graphics.setColor(0.3,0.3,0.3, Current_key_alpha)
  love.graphics.rectangle('line', x, y, w, 30, 5,5)
  love.graphics.print(Current_key, x+5, y+2)
end
