local utf8 = require 'utf8'

local array = require 'array'
local my_utf8 = require 'my_utf8'
local t = require 'utils'
local rects = require 'rects'
local Loc = require 'loc'

local move = {}
local I = {}
move.internal = I

---- moving around the editor in terms of locations (loc)
--
-- locations within text lines look like this:
--   {line=, pos=}
-- (pos counts in utf8 codepoints starting from 1)
--
-- all movements are built primarily using the following primitives, defined
-- further down:
--   - to_loc: (x, y) -> loc
--     identify the location at pixel coordinates (x,y) on screen
--     returns nil if (x,y) is not on screen
--   - to_coord: loc -> x, y
--     identify the top-left coordinate on screen of location loc
--     returns nil if loc is not on screen
--   - loc_down: loc, dy -> loc
--     find the location at the start of a screen line dy pixels down from loc
--     returns nil if dy is taller than the screen
--     returns bottom of file if we hit it
--   - loc_up: loc, dy -> loc
--     find the location at the start of a screen line dy pixels up from loc
--     returns nil if dy is taller than the screen
--     returns top of file if we hit it
--   - loc_hor: loc, x -> loc
--     find the location at x=x0 on the same screen line as loc
--
-- I have tried to make these definitions as clear as possible while being fast enough.
-- My mental model for trading off performance for clarity:
--   - any computation limited to the number of characters a screen can show
--     will be an order of magnitude faster than it needs to be to draw 30
--     frames per second.
--     So I don't mind doing up to 5 scans per interactive operation if that
--     makes the code clearer.
--   - no caching across frames; it makes the code less clear and has also
--     caused bugs.
--   - short-lived memory allocations that live within a single frame are cheap
--
-- This API was non-trivial to arrive at. It seems promising for any
-- pixel-based editor using proportional fonts. It seems independent of the
-- data structures the editor uses, whether arrays as here or ropes or gap
-- buffers, though I cheat a bit and use knowledge of the data structures
-- where it saves me a scan or two.

function move.up_arrow(editor)
  editor.screen_top = I.loc_up(editor, editor.screen_top, 1 --[[px]])  -- scan
  assert(editor.screen_top)
end  -- 1 scan

function move.down_arrow(editor)
  local dy0 = rects.height_of_screen_line(editor, editor.screen_top)
  editor.screen_top = I.loc_down(editor, editor.screen_top, dy0)  -- scan
  assert(editor.screen_top)
end  -- 1 scan

function move.pageup(editor)
  editor.screen_top = I.loc_up(editor, editor.screen_top, editor.bottom - editor.top - editor.line_height) -- scan
  assert(editor.screen_top)
end  -- 1 scan

function move.pagedown(editor)
  editor.screen_top = I.loc_down(editor, editor.screen_top, editor.bottom - editor.top - editor.line_height)  -- scan
  assert(editor.screen_top)
end  -- 1 scan

function move.maybe_snap_cursor_to_top_of_screen(editor)
  local _, y = I.to_coord(editor, editor.cursor)  -- scan
  if y == nil then
    editor.screen_top = I.loc_hor(editor, editor.cursor, editor.left)  -- scan
    assert(editor.screen_top)
  end
end  -- 1-2 scans

function move.maybe_snap_cursor_to_bottom_of_screen(editor)
  local _, y = I.to_coord(editor, editor.cursor)  -- scan
  local h = rects.height_of_screen_line(editor, editor.cursor)
  if y == nil or y > editor.bottom - editor.top - h then
    editor.screen_top = I.loc_up_whole_screen_lines(editor, editor.cursor, editor.bottom - editor.top - h)  -- scan
    assert(editor.screen_top)
  else
    -- no need to scroll
    return
  end
end  -- 1-2 scans

---- helpers for moving around the editor
-- These simulate drawing on screen without actually doing so.
--
-- These _do_ depend on the data structures the editor uses. But you should go
-- a long way just by reimplementing them if you ever change the data structures.

-- These are based on the output of rects.compute, which provides:
--  - an array of rects {x,y, dx,dy}, one for each line
--  - if it's a text line:
--    - rects for each screen line in rect[].screen_line_rects, and
--    - rects for each utf8 codepoint in rect[].screen_line_rects[].char_rects

-- return the location corresponding to a pixel
function move.to_loc(editor, mx,my)
  if my < editor.top then
    return t.deepcopy(editor.screen_top)
  end
  local x, y = mx - editor.left, my - editor.top
  local rect
  for line_index,line in array.each(editor.lines, editor.screen_top.line) do
    if line_index == editor.screen_top.line then
      rect = rects.compute(editor, editor.screen_top)
    else
      rect = rects.compute(editor, {line=line_index, pos=1})
    end
    if y < rect.dy then
      local s = I.find_y(rect.screen_line_rects, y)
      local char_rect = I.find_xy(s.char_rects, x, y)
      assert(char_rect)
      return {line=line_index, pos=char_rect.pos}
    else
      y = y - rect.dy
      if y <= 0 then
        break
      end
    end
    if y <= 0 then
      break
    end
  end
  -- below all lines; return final rect on screen
  local s = rect.screen_line_rects
  local c = s[#s].char_rects
  return {line=rect.line_index, pos=c[#c].pos}
end

function I.to_coord(editor, loc)  -- scans
  if Loc.lt(loc, editor.screen_top) then return end
  local y = editor.top
  for line_index, line in array.each(editor.lines, editor.screen_top.line) do
    local rect
    if line_index == editor.screen_top.line then
      rect = rects.compute(editor, editor.screen_top)
    else
      rect = rects.compute(editor, {line=line_index, pos=1})
    end
    if line_index == loc.line then
      for _,s in ipairs(rect.screen_line_rects) do
        for _,c in ipairs(s.char_rects) do
          if c.pos == loc.pos and c.show_cursor then
            return editor.left + c.x, y + c.y
          end
        end
      end
      assert(false, ('to_coord: invalid pos in text loc (%s,%s)'):format(loc.line, loc.pos))
    end
    y = y + rect.dy
    if y > editor.bottom then
      break
    end
  end
end  -- 1 scan

-- find the location at the start of a screen line dy pixels down from loc
-- return nil if dy is more than screen height away
-- return bottommost screen line in file if we hit it
function I.loc_down(editor, loc, dy)  -- scans
  local y = 0
  local prevloc = loc
  for line_index, line in array.each(editor.lines, loc.line) do
    local rect = rects.compute(editor, {line=line_index, pos=1})
    for _,s in ipairs(rect.screen_line_rects) do
      if line_index > loc.line or loc.pos < s.pos+s.dpos then
        local currloc = {line=line_index, pos=s.pos}
        if y + s.dy > dy then
          return currloc
        end
        y = y + s.dy
        prevloc = currloc
      end
    end
  end
  return prevloc
end  -- 1 scan

-- find the location at the start of a screen line dy pixels up from loc
-- return nil if dy is more than screen height away
-- return topmost screen line in file if we hit it
function I.loc_up(editor, loc, dy)  -- scans
  local y = 0
  -- special handling for loc's line
  local rect = rects.compute(editor, {line=loc.line, pos=1})
  local found = false
  for is = #rect.screen_line_rects,1,-1 do
    local s = rect.screen_line_rects[is]
    if not found and I.within(loc.pos, s.pos, s.pos+s.dpos) then
      found = true
    elseif found then
      if y + s.dy > dy then
        return {line=loc.line, pos=s.pos}
      end
      y = y + s.dy
    else
      -- below loc's screen line; skip
    end
  end
  for line_index = loc.line-1,1,-1 do
    local line = editor.lines[line_index]
    local rect = rects.compute(editor, {line=line_index, pos=1})
    for is = #rect.screen_line_rects,1,-1 do
      local s = rect.screen_line_rects[is]
      if y + s.dy > dy then
        return {line=line_index, pos=s.pos}
      end
      y = y + s.dy
    end
  end
  return I.top_loc(editor)
end  -- 1 scan

-- find the location at the start of a screen line up to dy pixels up from loc, but count only whole screen lines.
-- So the result will be at or below dy pixels above loc.
function I.loc_up_whole_screen_lines(editor, loc, dy)  -- scans
  local prevloc = loc  -- bug: not at start of screen line
  local y = 0
  -- special handling for loc's line
  local rect = rects.compute(editor, {line=loc.line, pos=1})
  local found = false
  for is = #rect.screen_line_rects,1,-1 do
    local s = rect.screen_line_rects[is]
    if not found and I.within(loc.pos, s.pos, s.pos+s.dpos) then
      found = true
    elseif found then
      local currloc = {line=loc.line, pos=s.pos}
      if y + s.dy > dy then
        return prevloc
      end
      y = y + s.dy
      prevloc = currloc
    else
      -- below loc's screen line; skip
    end
  end
  for line_index = loc.line-1,1,-1 do
    local line = editor.lines[line_index]
    local rect = rects.compute(editor, {line=line_index, pos=1})
    for is = #rect.screen_line_rects,1,-1 do
      local s = rect.screen_line_rects[is]
      local currloc = {line=line_index, pos=s.pos}
      if y + s.dy > dy then
        return prevloc
      end
      y = y + s.dy
      prevloc = currloc
    end
  end
  return I.top_loc(editor)
end  -- 1 scan

function I.top_loc(editor)
  assert(#editor.lines > 0)
  return {line=1, pos=1}
end

-- find the location at x=x0 on the same screen line as loc
function I.loc_hor(editor, loc, x0)  -- scans line
  local rect = rects.compute(editor, {line=loc.line, pos=1})
  x0 = x0 - editor.left
  assert(rect.screen_line_rects)
  for i,sc in ipairs(rect.screen_line_rects) do
    if loc.pos >= sc.pos and loc.pos < sc.pos+sc.dpos then
      local prevx = nil
      for _,c in ipairs(sc.char_rects) do
        if (prevx == nil or x0 >= (prevx+c.x)/2) and x0 < c.x+c.dx/2 and c.show_cursor then
          return {line=loc.line, pos=c.pos}
        end
        prevx = c.x
      end
      return {line=loc.line, pos=sc.pos+sc.dpos-1}
    end
  end
end  -- 0-1 scans

function I.find_xy(rects, x, y)
  if x < 0 then return rects[1] end
  for _, rect in ipairs(rects) do
    if I.within_rect(rect, x, y) then
      return rect
    end
  end
  return rects[#rects]
end

function I.find_x(rects, x)
  if x < 0 then return rects[1] end
  for _, rect in ipairs(rects) do
    if I.within(x, rect.x, rect.x+rect.dx) then
      return rect
    end
  end
  return rects[#rects]
end

function I.find_y(rects, y)
  if y < 0 then return rects[1] end
  for _, rect in ipairs(rects) do
    if I.within(y, rect.y, rect.y+rect.dy) then
      return rect
    end
  end
  return rects[#rects]
end

function I.within_rect(rect, x,y)
  return I.within(x, rect.x, rect.x+rect.dx)
    and I.within(y, rect.y, rect.y+rect.dy)
end

function I.within(a, lo, hi)
  return a >= lo and a < hi
end

return move
