local utf8 = require 'utf8'

local my_utf8 = require 'my_utf8'
local my_string = require 'my_string'
local my_table = require 'my_table'
local t = require 'utils'

local prose = {}
local I = {}
prose.internal = I

function prose.compute_rects(editor, loc)
  local subtle_color = require 'subtle_color'
  local line = editor.lines[loc.line]
  local blanks = I.special_ranges(line)
  local screen_lines = {}
  local curr_screen_line = {}
  local spos = loc.pos
  local x, y = 0, 0
  local dy = editor.line_height  -- the height of the current screen line
  local curr_blank = I.special_range_after(blanks, loc.pos)
  for pos,char in my_utf8.chars(line, loc.pos) do
    if curr_blank > #blanks or pos <= blanks[curr_blank].s then
      -- regular prose
      local w = editor.font:getWidth(char)
      if char:match('%s') then
        if I.line_wrap_at_word_boundary(editor, editor.left + x, line, pos) then
          table.insert(curr_screen_line,
            {x=x, y=y, dx=w, dy=editor.line_height, pos=pos, data=char, show_cursor=true})  -- char
          table.insert(curr_screen_line,
            {x=x+w, y=y, dx=editor.width-x-w, dy=editor.line_height, pos=pos+1})  -- filler
          table.insert(screen_lines,
            {x=0, y=y, dx=editor.width, dy=dy,
             pos=spos, dpos=pos-spos+1, char_rects=curr_screen_line})
          curr_screen_line = {}
          spos = pos+1
          x = 0
          y = y + dy
          dy = editor.line_height
          if y > editor.bottom - editor.top then
            return {x=0, y=0, dx=editor.width, dy=y, screen_line_rects=screen_lines}
          end
        else
          table.insert(curr_screen_line,
            {x=x, y=y, dx=w, dy=editor.line_height, pos=pos, data=char, show_cursor=true})
          x = x + w
        end
      else
        if x+w > editor.width then
          assert(pos > 1)
          table.insert(curr_screen_line,
            {x=x, y=y, dx=editor.width-x, dy=editor.line_height, pos=pos})  -- filler
          table.insert(screen_lines,
            {x=0, y=y, dx=editor.width, dy=editor.line_height,
             pos=spos, dpos=(pos-1)-spos+1, char_rects=curr_screen_line})
          curr_screen_line = {}
          spos = pos
          x = 0
          y = y + dy
          dy = editor.line_height
          if y > editor.bottom - editor.top then
            return {x=0, y=0, dx=editor.width, dy=y, screen_line_rects=screen_lines}
          end
        else
          -- nothing
        end
        local style = {}
        if (curr_blank <= #blanks and pos == blanks[curr_blank].s)
          or (curr_blank > 1 and pos == blanks[curr_blank-1].e)
        then
          -- leave out markup characters
        else
          table.insert(curr_screen_line,
            {x=x, y=y, dx=w, dy=editor.line_height, pos=pos, data=char, show_cursor=true})
          x = x + w
        end
      end
    else
      -- blank
      local s, p, e = blanks[curr_blank].s+1, blanks[curr_blank].p, blanks[curr_blank].e-1
      assert(s <= p and p <= e)
      if curr_blank <= #blanks and pos == s then
        local numerator = my_utf8.substring(line, s, p-1)
        local denominator = my_utf8.substring(line, p+1, e)
        local style = {}
        if blanks[curr_blank].type == '<' then
          style = I.parse_style(denominator)
        end
        local numw = t.width(numerator)
        assert(numw <= editor.width, 'wrapping within a blank is not yet implemented')
        local w = numw
        if x+w > editor.width then
          table.insert(screen_lines,
            {x=0, y=y, dx=editor.width, dy=dy,
              pos=spos, dpos=pos-spos+1, char_rects=curr_screen_line})
          curr_screen_line = {}
          x = 0
          y = y + dy
        end
        -- numerator
        local save_x = x
        local padding = (w - numw)/2
        assert(padding == 0)
        for _,char in my_utf8.chars(numerator) do
          local w = t.width(char)
          table.insert(curr_screen_line,
            my_table.fill_defaults(
              {x=x, y=y, dx=w, dy=editor.line_height, pos=pos, data=char, show_cursor=true},
              style))
          x = x + editor.font:getWidth(char)
          pos = pos+1
        end
        -- horizontal line
        pos = pos+1
        -- denominator
        pos = pos+utf8.len(denominator)
      elseif curr_blank <= #blanks and pos > s and pos < e then
        -- already processed
      elseif curr_blank <= #blanks and pos == e then
        curr_blank = curr_blank+1
      end
    end
  end
  table.insert(curr_screen_line,
    {x=x, y=y, dx=editor.width-x, dy=editor.line_height, pos=utf8.len(line)+1, show_cursor=true})  -- filler
  table.insert(screen_lines,
    {x=0, y=y, dx=editor.width, dy=dy,
     pos=spos, dpos=utf8.len(line)+1-spos+1, char_rects=curr_screen_line})
  y = y + dy
  return {x=0, y=0, dx=editor.width, dy=y, line_index=loc.line, screen_line_rects=screen_lines}
end

function I.special_range_after(ranges, pos)
  for i,range in ipairs(ranges) do
    if range.s >= pos then
      return i
    end
  end
  return #ranges+1
end

function prose.height_of_screen_line(editor, loc)
  local line = editor.lines[loc.line]
  local rects = prose.compute_rects(editor, {line=loc.line, pos=1})
  for _,s in ipairs(rects.screen_line_rects) do
    if loc.pos >= s.pos and loc.pos <= s.pos+s.dpos then
      return s.dy
    end
  end
end

function I.parse_style(s)
  local result = {}
  for k,v in s:gmatch('(%S+)=(%S+)') do
    local v2 = my_string.split(v, ',')
    if k == 'fg' or k == 'bg' then
      if #v2 >= 3 and tonumber(v2[1]) and tonumber(v2[2]) and tonumber(v2[3]) then  -- wait until we've got a valid color
        result[k] = {r=v2[1], g=v2[2], b=v2[3]}
      end
    elseif #v2 > 1 then
      result[k] = v2
    else
      result[k] = v
    end
  end
  return result
end

-- helper to check if space character at pos, x is final possible candidate
-- for line wrapping before x = editor.right
-- x lies between editor.left and editor.right
function I.line_wrap_at_word_boundary(editor, x, line, pos)
  if x - editor.left > 0.8 * (editor.right - editor.left) then
    -- word boundary more than 80% of the way across.
    -- check if there's any more word boundaries before x gets to editor.right
    for pos, char in my_utf8.chars(line, pos+1) do
      if x > editor.right then
        return true
      end
      local w = editor.font:getWidth(char)  -- width of char
      if char:match('%s') then
        if x+w < editor.right then
          break
        end
      end
      x = x+w
    end
  end
end

-- return a table {s[tart]=, p[ipe]=, e[nd]=} pos, counting in utf-8 codepoints
-- corresponding to the '[', '|', ']' for each blank,
-- or the '<', '|', '>' for each inline style.
-- Don't put '[' or ']' inside styles, or '<' or '>' inside blanks. If we ever
-- need that we'll need a whole new notation.
function I.special_ranges(line)
  local result = {}
  I.append_all_ranges(line, '<[^%|<>]*%|[^%]<>]*>', result)  -- styles
  return result
end

-- append all {start, pipe, end} occurrences in line of pattern to output
-- pattern must contain a single '|'
function I.append_all_ranges(line, pattern, output)
  local max_pos = utf8.len(line)
  local curr_byte, curr_pos = 1, 1
  while curr_byte do
    local s, e = line:find(pattern, curr_byte)
    if s == nil then break end
--?     local spos, epos = my_utf8.pos(line, s), my_utf8.pos(line, e)
    local p = line:find('|', s)
    local spos = my_utf8.pos(line, s, curr_pos, max_pos)
    local ppos = my_utf8.pos(line, p, curr_pos, max_pos)
    local epos = my_utf8.pos(line, e, spos, max_pos)
    table.insert(output, {s=spos, p=ppos, e=epos, type=pattern:sub(1,1)})
    curr_byte, curr_pos = e, epos
  end
end

return prose
