local utf8 = require 'utf8'

local prose = require 'prose'

local rects = {}
local I = {}
rects.internal = I

rects.initialize_sections = prose.initialize_sections

-- Generate rects for each screen line in it and the range [pos,pos+dpos-1]
-- associated with each within each screen line generate rects for each char
-- (utf8 codepoint) and the pos associated with each.
--
-- Each rect is a rectangle on screen, as defined by its x, y, dx (width) and
-- dy (height). A rect will also usually contain some data that is
-- claimed/made to live within that rectangle on screen.
--
-- Example line containing 3 screen lines after word wrapping:
--  {x=0, y=0, dx=100, dy=30,  -- line is 100px wide and 30px tall
--    screen_line_rects = {
--      {x=0, y=0, dx=100, dy=10,  -- first screen line of line
--        ...  -- see below for what's inside a screen line
--      },
--      {x=0, y=10, dx=100, dy=10,  -- second screen line of line starts at y=10
--        ...
--      },
--      {x=0, y=20, dx=100, dy=10,  -- third screen line of line starts at y=20
--        ...
--      },
--    },
--  }
--
-- Example screen line:
--  {x=0, y=0, dx=100, dy=10,
--    pos=1, dpos=10,  -- will render 10 characters from the line starting at position 1 (start of line)
--    char_rects = {
--      {x=0, y=0, dx=10, dy=10, pos=1, data='a'},  -- the first character is drawn at 0,0; clicking anywhere in this rect focuses cursor before this character
--      ...
--    }
--  }
--
-- Some experimental features for syntax highlighting or "syntax geometry":
--
-- * A char rect can only show the cursor inside it if show_cursor is set.
--   This enables padding characters. You can have a range of rects on screen
--   that all map to the same location (only one of them sets data). Clicking
--   on any of them positions cursor in the same location. But no matter where
--   you click, the cursor shows in the same place.
--
--   Example:
--    char_rects = {
--      {x=0, y=0, dx=10, dy=10, pos=1, data='a', show_cursor=true},  -- now you need this extra verbosity by default
--      ...
--    }
--
-- * The 'data' in a char rect won't be drawn if 'conceal' is set. The rect
--   still exists and might position the cursor if you click on it. But what
--   would it look like? That brings me to:
--
-- * A char rect can include arbitrary 'draw' commands. For now I only support
--   a couple: lines and rectangles.
--
--   Example: draw an 'a' on screen with a border around it
--    char_rects = {
--      {x=0, y=0, dx=10, dy=10, pos=1, data='a', show_cursor=true
--        draw = {
--          {type='rect', mode='line', x=0, y=0, w=10, h=10},
--        },
--      },
--      ...
--    }
--
--   Example: draw an 'a' on screen with an underline
--    char_rects = {
--      {x=0, y=0, dx=10, dy=10, pos=1, data='a', show_cursor=true
--        draw = {
--          {type='line', x1=0, y=10, x2=10, y2=10)
--        },
--      },
--      ...
--    }
--
-- Influencing where to show the cursor, when to hide text, or when show
-- graphics in addition to or instead of the text, these seem like they might
-- cover everything we need from syntax highlighting or "syntax geometry".
function rects.compute(editor, loc)
  return prose.compute_rects(editor, loc)
end

function rects.height_of_screen_line(editor, loc)
  return prose.height_of_screen_line(editor, loc)
end

return rects
